"""
File Purpose: MainDimensionsChunker

Manages "chunks" options for main dimensions (see also: MainDimensionsHaver).

See docstring for MainDimensionsChunker for details.
See also: https://gitlab.com/Sevans7/plasmacalcs/-/issues/37

--- CAUTION ---
    derivative results different with/without chunks, unless deriv_before_slice=True.
    All other results should be the same with/without chunking.
"""

import xarray as xr

from .main_dimensions import MainDimensionsHaver
from ..errors import ChunkError, ChunkDimensionalityError
from ..tools import (
    UNSET,
    simple_property,
    TaskList,
)


class Chunker():
    '''manages choosing slices based on chunks.

    --- CAUTION ---
        derivative results different with/without chunks, unless deriv_before_slice=True.
        All other results should be the same with/without chunking.

    chunks: dict
        keys should be maindim possibly with '_size' or '_slices' appended, e.g. 'x' or 'y_size'.
        x --> break x into this many chunks total.
        x_size --> break x into chunks of this size.
        x_slices --> list of slices to use for x.
    maindims_sizes: dict
        {dim: full size} for dims in maindims.
    init_slices: None or dict
        dict of slices which already exist.
        un-chunked dims' slices will be maintained when creating chunk slices.
        invalid chunking if any chunk dims appear in `slices`.
        None --> internally store as empty dict.
    lenient: bool
        whether to allow any ChunkSlicer inputs, even if invalid.
        default False, i.e. self.check_valid() (which will raise ChunkError if not valid).
    '''
    def __init__(self, chunks, maindims_sizes, init_slices=None, *, lenient=False):
        self.chunks = chunks
        self.maindims_sizes = maindims_sizes
        self.init_slices = {} if init_slices is None else init_slices
        if not lenient:
            self.check_valid()

    def __repr__(self):
        return f'{type(self).__name__}({self.chunks}, maindims_sizes={self.maindims_sizes})'

    @property
    def chunk_dims(self):
        '''the tuple of dims which are being chunked.
        E.g., if chunks={'x':10, 'y_size':50}, then chunking_dims='x', 'y'.
        Computed as set of self.chunks keys, with '_size' or '_slices' removed if present.
        '''
        result = tuple({dim[:-len('_slices')] if dim.endswith('_slices') else
                        dim[:-len('_size')] if dim.endswith('_size') else dim
                        for dim in self.chunks})
        return result

    @property
    def maindims(self):
        '''the maindims. Just the tuple of self.maindims_sizes keys.'''
        return tuple(self.maindims_sizes)

    def check_valid(self):
        '''checks that self is a valid ChunkSlicer. crash with ChunkError if not.
        Valid must satisfy:
            - all chunk dims are maindims
            - no overlap between chunks and slices (e.g. 'x' in chunks and slices)
        '''
        for x in self.chunk_dims:
            if x not in self.maindims:
                raise ChunkError(f'Invalid chunk dim {x!r}; not found in maindims={self.maindims}')
            if x in self.init_slices:
                raise ChunkError(f'Cannot chunk dim {x!r}; it is already being sliced: slices={self.init_slices[x]!r}')

    def dim_chunk_info(self, x):
        '''return chunking info for dim x: (mode, value).
        mode will be 'n', 'size', or 'slices',
        correspond to x, x_size, or x_slices appearing in self.chunks.
        value will be int, int, or iterable of slices, respectively.
        '''
        if x not in self.chunk_dims:
            raise ChunkError(f'Invalid chunk dim {x!r}; not found in self.chunks={self.chunks}')
        if x in self.chunks:
            return 'n', self.chunks[x]
        elif f'{x}_size' in self.chunks:
            return 'size', self.chunks[f'{x}_size']
        elif f'{x}_slices' in self.chunks:
            return 'slices', self.chunks[f'{x}_slices']

    def _iter_slicers_x(self, x):
        '''iterates through all chunk slices for dim x.'''
        mode, chunk = self.dim_chunk_info(x)
        full_size = self.maindims_sizes[x]
        if mode == 'n' or mode == 'size':
            if mode == 'size':
                x_size = chunk
            else:
                n = chunk
                x_size = (full_size // n) + (0 if full_size % n == 0 else 1)
            for x_start in range(0, full_size, x_size):
                x_stop = min(x_start + x_size, full_size)
                yield slice(x_start, x_stop)
        elif mode == 'slices':
            for s in chunk:
                yield s
        else:
            raise ChunkError(f'Invalid chunking mode {mode!r} for dim {x!r}')

    def iter_slicers(self):
        '''iterates through all chunk slices, yielding slices dict including any non-chunked slices.

        E.g. chunks=dict(x_size=50), slices=dict(z=7) --> yield:
            dict(x=slice(0,50), z=7),
            dict(x=slice(50,100), z=7),
            ...
        '''
        chunk_dims = self.chunk_dims
        if len(chunk_dims) == 0:
            # no chunking; just yield initial slices once
            yield self.init_slices
        elif len(chunk_dims) == 1:
            # single chunk dim; simple loop
            x = chunk_dims[0]
            for xslice in self._iter_slicers_x(x):
                slices = self.init_slices.copy()
                slices[x] = xslice
                yield slices
        elif len(chunk_dims) == 2:
            # two chunk dims, also sort of simple to loop:
            x, y = chunk_dims
            for xslice in self._iter_slicers_x(x):
                for yslice in self._iter_slicers_x(y):
                    slices = self.init_slices.copy()
                    slices[x] = xslice
                    slices[y] = yslice
                    yield slices
        else:
            # could easily extend pattern to 3D+, but should generalize if doing that.
            # but, probably not necessary!
            raise NotImplementedError('[TODO] chunking along 3 or more dims')

    def slicers(self):
        '''list of all slices to use for chunking. Equivalent: list(self.iter_slicers()).'''
        return list(self.iter_slicers())

    def __len__(self):
        '''the number of chunk slices which will be generated by self.iter_slicers().
        [EFF] perhaps should make this more efficient instead of generating all slices...
        '''
        return len(self.slicers())


class MainDimensionsChunker(MainDimensionsHaver):
    '''Manages "chunks" options for main dimensions.'

    --- CAUTION ---
        derivative results different with/without chunks, unless deriv_before_slice=True.
        All other results should be the same with/without chunking.

    Here is the idea behind chunks:
        For data with large arrays along maindims, sometimes arrays are
        so large that loading the whole thing into memory slows down computations a lot.
        Also, some computations are local, depending on no other cells or only nearby cells.

        In those cases, it is probably much faster to break up the maindims into chunks,
        load one chunk at a time, and combine results appropriately.

    The internal implementation goes something like this:
        - chunks will be a behavior attribute (set it via self.chunks=val, or sel(var, chunks=val)).
        - if chunks is set, hijack the self.__call__ to instead do chunking.
        - doing chunking means:
            - determining a list of appropriate slices along maindims, based on chunks
            - getting results for each slice (while getting those results, disable chunking)
            - combining the results appropriately

    Currently, chunks can only apply to vars which keep the chunked axis or axes.
    So, for example, if maindims=('x', 'y', 'z') and 'u' has spatial extent:
        - self('mod_u') can be chunked
        - self('mean_u') cannot be chunked;
            mean_var takes the mean across maindims, so result has no spatial extent.
        - self('fft_u') cannot be chunked;
            fft_var converts dims (default: fft_dims=maindims) into frequency dims.
    Internally, there is no check ahead of time to see if a var can be chunked.
    Instead, chunking will be attempted, and then will check if the chunked axes remain;
        if not, will crash with ChunkDimensionalityError.
    '''

    # # # CHUNKS DEFINITIONS # # #
    chunker_cls = Chunker

    cls_behavior_attrs.register('chunks', default={})
    chunks = simple_property('_chunks', setdefault=dict,
        doc='''chunks for maindims when loading arrays.
        E.g. chunks = dict(x=10, y_size=50)
            --> break x into 10 chunks total, and break y into chunks of size 50.
                (if y not divisible by 50, the last chunk will be smaller.)
        Can also provide a list of indexers for more low-level control,
        E.g. chunks = dict(x=[slice(0,50), 100, slice(800,900)])
            --> repeat calculation with x from 0 to 50, then 100th value, then 800 to 900.
            result will be a well-labeled xarray with x at those coordinates,
            here equivalent to result from slices=dict(x=[0,1,...,49,100,800,801,...,899])
        See help(self.chunker_cls) or help(self.chunker()) for more details.

        Notes:
            - mutually exclusive with providing slices of the same dims.
                E.g., cannot provide x chunk and x slice at the same time.
            - if provided, internally hijacks self.__call__ to do chunking instead,
                and inside of that will loop through calls to self with different slices.
            - only works for vars which keep the chunked axes.
                E.g., fails for self('mean_n') since mean result doesn't have maindims.''')
    @chunks.setter
    def chunks(self, value):
        if value is None:
            value = dict()
        self._chunks = value

    def chunker(self):
        '''gets Chunker based on the current values of self.chunks and self.slices.
        see help on MainDimensionsChunker, or help(type(self).chunks) for more details.
        '''
        return self.chunker_cls(self.chunks, self.maindims_full_sizes, self.slices)

    # # # CALLING SELF, WITH CHUNKS # # #
    def _call_hijacker(self, *args_super, **kw_super):
        '''returns False or name of hijacker method to use instead of self(var) call.
        Here, returns '_get_with_chunks' if self.chunks is nonempty, else super()._call_hijacker(...).
        '''
        if len(self.chunks) > 0:
            return '_get_with_chunks'
        else:
            return super()._call_hijacker(*args_super, **kw_super)

    def _get_with_chunks(self, var, *args_super, **kw_super):
        '''returns value of var, but chunking while loading, then merging chunked results.
        Result should be equivalent to self(var, chunks=None).
        CAUTION: derivatives may give wrong answers at chunk boundaries, unless deriv_before_slices=False
        '''
        chunker = self.chunker()
        chunkdims = chunker.chunk_dims
        if len(chunkdims) != 1:
            raise NotImplementedError('[TODO] implement chunking when len(chunk_dims) != 1')
        ## IMPLEMENTATION 1: naive (single-dim only, and no multiprocessing) implementation
        # (implementation 1 added in commit 55ac709. Removed since then.)

        ## IMPLEMENTATION 2: slightly less naive (single-dim only, but includes multiprocessing):
        chunkdim = chunkdims[0]
        slicers = chunker.slicers()
        task_inputs = [
            (self, (var, *args_super), {**kw_super, 'slices': slicer})
            for slicer in slicers
        ]
        task_inputs[0][2]['_chunks_check0'] = (chunkdim, self.call_depth + 1)  # will check if chunking allowed
        tasks = TaskList(*task_inputs)
        mp_strategy = dict(ncpu=self.ncpu,   # get multiprocessing preferences from self
                   timeout=self.timeout,
                   ncoarse=self.ncoarse,
                   print_freq=self.print_freq_explicit)
        with self.using(chunks=None, ncpu=1):  # internally, don't use chunking nor multiprocessing.
            results = tasks(**mp_strategy)
        return xr.concat(results, dim=chunkdim)

    # # # CHUNK CHECKS (BOOKKEEPING) # # #
    @property
    def _extra_kw_for_quantity_loader_call(self):
        '''extra kwargs which can be used to set attrs self during self.__call__.
        The implementation here returns ['_chunks_check0'] + any values from super().
        '''
        return ['_chunks_check0'] + super()._extra_kw_for_quantity_loader_call

    _chunks_check0 = simple_property('_chunks_check0_', default=None,
        doc='''None or (dim, call_depth) --> assert self(...) first chunk has dim at this call depth.
        E.g., _chunks_check0 = ('x', 3) --> assert 'x' in result.dims at call_depth=3,
            and if it isn't, raise ChunkDimensionalityError.''')

    def _call_postprocess(self, result, *, var, name=UNSET, item=UNSET):
        '''postprocess result from self.__call__. Called during self.__call__.
        (self.call_depth inside here will tell depth of the current call; depth=1 for top level.)

        result: any value, probably an xarray.DataArray
            result from self.__call__, before postprocessing.
        var, name, item: UNSET or value
            passed directly from self.__call__.

        The implementation here does the following (subclasses might override / add to this):
            (A) call super()._call_postprocess(), which probably (if parent is QuantityLoader) does:
                (1) if self.verbose >= 4, print a message about getting var.
                (2) result = self.attach_extra_coords(result).
                (3) set result.name = name, or result.attrs['name'] = name, if possible.
                (4) if self.assign_behavior_attrs at this call depth, do so now.
                (5) if self.call_depth == 1, call self._call_postprocess_toplevel.
                (6) if item, convert result to single value via result.item().
            (B) if self._chunks_check0 is not None, crash if chunkdim is missing from result.dims.
                else, return result, unchanged.
        '''
        result = super()._call_postprocess(result, var=var, name=name, item=item)
        if self._chunks_check0 is not None:
            x, depth = self._chunks_check0
            # depth is needed so that we don't do the checks internally (at larger depths),
            #   e.g. imagine var='5*n', we shouldn't do the check for '5', only for '5*n'.
            if self.call_depth == depth:
                if x not in getattr(result, 'dims', []):
                    errmsg = (f"Cannot chunk {var!r} along {x!r}. "
                              f"getting var {var!r} removes dim {x!r}; "
                              f"first chunk's result had dims={result.dims})")
                    raise ChunkDimensionalityError(errmsg)
        return result



"""
File Purpose: LoiLoader

LOI stands for "Line-Of-Integration", or if you prefer,
    "Lots Of Interpolation", "Let's dO Interpolation", or "Let's do Our Interpolation".

It is useful for loading quantities when interpolating onto a new set of coordinates,
    e.g. a path (or multiple paths) through a box, then maybe integrating along the path.
"""

from .vector_arithmetic import VectorArithmeticLoader
from ...errors import (
    InputError, InputMissingError, InputConflictError,
    FormulaMissingError,
    ComponentValueError,
)
from ...tools import (
    UNSET,
    alias, simple_property, format_docstring,
    xarray_d_grid,
)


class LoiLoader(VectorArithmeticLoader):
    '''Line-Of-Integration (LOI) loader, for interpolating onto a new set of coordinates,
    e.g. a path (or multiple paths) through a box, then maybe integrating along the path.
    '''
    cls_behavior_attrs.register('loi', default='los')
    loi = simple_property('_loi', default='los',
        doc='''Line Of Integration (dim, dict of arrays for interp), or "Lots Of Interpolation", if you prefer.
        Used by self("var_loi") pattern (for the interpolation) and any quantities which integrate along a line.
        ([TODO] move the integrate-along-a-line patterns here, e.g. from mhd_radiative_loader.)

        Can be any of the following:
            None --> no integrate-along-a-line (those methods will crash), and no interpolation.
                self("var_loi") == self("var").
            dict --> no integrate-along-a-line (those methods will crash), but can interpolate onto these arrays.
                self("var_loi") == self("var").interp(**self.loi)
            str --> special option (e.g. 'los' for "use behavior implied by self.los")
                see self.LOI_OPTIONS for all special options.
            2-tuple --> (integration instructions, interpolation instructions)
                integration instructions can be None (as above) or str telling dim to integrate along.
                    Note: if dim ends with '_dim', assumes dim[:-4] tells coord values,
                        e.g. x_dim implies integration along 'x_dim' but dx comes from 'x' coord,
                        which is a useful pattern if 'x' coord depends on multiple dims.
                    Note: if dim starts with '-', assumes loi_cumsum should be evaluated in reverse order.
                        Physically, this can affect things like optical depth.
                interpolation instructions can be None (as above) or dict (as above).

        (To supply kwargs to the interpolator, provide them in the dict.
            E.g.: dict(x=array(...), y=array(...), assume_sorted=True).
            Or, for clarity: dict(coords=(x=array(...), y=array(...)), assume_sorted=True).)''')

    LOI_OPTIONS = {
        'los': '''behavior implied by self.los: no interpolation; integration is along line-of-sight dimension.
            Integration only works if line-of-sight implies a dimension (e.g. 'x')
                (line-of-sight can generally be an array).''',
    }

    loi_explicit = alias('loi',
        doc='''loi as 2-tuple of (integration axis, None or dict of arrays for interpolation).
        Getting loi_explicit will crash if no implied integration axis.
        If integration axis starts with '-', it implies loi_cumsum() in reverse order.
            e.g. '-x' is just like 'x' but any loi cumsums will be in reverse.
            Physically, this can affect quantities like optical depth.''')
    @loi_explicit.getter
    def loi_explicit(self):
        idim = self._loi_integration_dim(signed=True)  # <-- crashes if appropriate
        interp = self._loi_interp_dict()  # <-- crashes if appropriate
        return idim, interp

    def _loi_interp_dict(self):
        '''return dict of arrays to interpolate onto, as implied by self.loi.
        result is None if self.loi implies "no interpolation".
        Crash if invalid self.loi.
        '''
        loi = self.loi
        if loi is None:
            result = None
        elif isinstance(loi, dict):
            result = loi
        elif isinstance(loi, str):
            if loi == 'los':
                result = None
            elif loi in self.LOI_OPTIONS:
                raise NotImplementedError(f'Valid LOI string option {loi!r}; subclass should implement.')
            else:
                raise InputError(f'Invalid LOI string option {loi!r}. Options: {list(self.LOI_OPTIONS.keys())}.')
        elif len(loi) == 2:
            result = loi[1]
        else:
            raise InputError(f'Invalid LOI value {loi!r}. See help(type(self).loi) for valid options.')
        return result

    def _loi_integration_dim(self, *, signed=False):
        '''return dimension along which to integrate, as implied by self.loi.
        Crash with InputMissingError if self.loi implies no integration.

        signed: bool
            whether to start with '-' to imply "cumsum in reverse order" inside loi_cumsum().
            E.g. '-z' --> array.isel(z=slice(None,None,-1)).cumsum('z').isel(z=slice(None,None,-1)
        '''
        loi = self.loi
        neg = False
        if loi is None or isinstance(loi, dict):
            raise InputMissingError(f'self.loi={loi!r} implies no integration; cannot get integration dim.')
        elif isinstance(loi, str):
            if loi == 'los':
                los = self.los
                if not isinstance(los, str):
                    raise InputConflictError(f'self.loi="los" requires string los, but got self.los={self.los!r}')
                if los.startswith('-'):
                    neg = True
                    los = los[1:]
                if los in ('x', 'y', 'z'):
                    result = los
                elif los == 'component':
                    if self.current_n_component() == 1:
                        result = str(self.component)
                    else:
                        errmsg = (f'When LOI option "los" implies integration along self.los == {self.los!r}, '
                                  f'need self.current_n_component()==1, but got self.component={self.component!r}.')
                        raise ComponentValueError(errmsg)
                else:
                    raise NotImplementedError(f'self.loi="los" when self.los={self.los!r}')
            else:
                raise InputError(f'Invalid LOI string option {loi!r}. Options: {list(self.LOI_OPTIONS.keys())}.')
        elif len(loi) == 2:
            if loi[0] is None:
                raise InputMissingError(f'self.loi={loi!r} implies no integration; cannot get integration dim.')
            elif isinstance(loi[0], str):
                if loi[0].startswith('-'):
                    neg = True
                    result = loi[0][1:]
            else:
                raise InputError(f'Invalid self.loi[0]. Expected None or str, got {loi[0]!r}.')
        else:
            raise InputError(f'Invalid LOI value {loi!r}. See help(type(self).loi) for valid options.')
        if signed and neg:
            result = '-' + result
        return result

    def _loi_integration_coord(self):
        '''return coord to use for integration, as implied by self.loi.
        Just like self._loi_integration_dim(), but if result ends with '_dim', remove it.
        E.g. if _loi_integration_dim() == 'grid_dim', this returns 'grid'.
        '''
        idim = self._loi_integration_dim()
        return idim[:-len('_dim')] if idim.endswith('_dim') else idim
    
    @known_pattern(r'(.+)_loi', deps=[0])
    def get_var_loi(self, var, *, _match=None):
        '''var_loi = var interpolated onto coordinates implied by self.loi.
        Internally calls self.loi_interp(val) to do the interpolation (if any);
            can reliably use that method for same functionality as array_loi pattern, anywhere.

        Like all kwarg/attrs, can set loi beforehand, or as kwarg inside the call:
            self.loi = value; self('u_loi')  # option 1
            self('u_loi', loi=value)  # option 2
        '''
        here, = _match.groups()
        arr = self(here)
        return self.loi_interp(arr)

    def loi_interp(self, array):
        '''return array interpolated onto coordinates implied by self.loi.
        See help(type(self).loi) for more details.
        Returns array unchanged if self.loi implies no interpolation.
        '''
        loi_interp_dict = self._loi_interp_dict()
        if loi_interp_dict is None:
            result = array
        else:
            result = array.interp(**loi_interp_dict)
        return result

    @known_pattern(r'loi_cumsum_(.+)', deps=[0])
    def get_loi_cumsum_var(self, var, *, _match=None):
        '''loi_cumsum_var = var interpolated onto coordinates implied by self.loi,
        then cumulatively summed along line-of-integration dimension implied by self.loi.
        Internally calls self.loi_cumsum(val); can reliably use that method instead if desired.
        '''
        here, = _match.groups()
        arr = self(here)
        return self.loi_cumsum(arr, do_interp=True)

    cls_behavior_attrs.register('loi_cumsum_sort', default=True)
    loi_cumsum_sort = simple_property('_loi_cumsum_sort', default=True,
        doc='''Tells whether to sort along the integration dimension before doing cumsum in self.loi_cumsum().
        (If sorting changes order, will not restore original order after cumsum operation,
            so operations relying on xarray coord matching would still work,
            but underlying data order will be different between result and original array.)
        True or 'ascending' --> sort in ascending order.
        'descending' --> sort in descending order.
        False --> do not sort.
        This check occurs before any reversing implied by self.loi.
        Example: self.loi_explicit[0] == '-z', loi_cumsum_sort=True
            --> array.sortby('z').isel(z=slice(None,None,-1)).cumsum('z').isel(z=slice(None,None,-1))''')

    @format_docstring(_sort_doc=loi_cumsum_sort.__doc__, sub_ntab=1)
    def loi_cumsum(self, array, *, do_interp=True, sort=UNSET):
        '''return array interpolated onto coordinates implied by self.loi,
        then cumulatively summed along line-of-integration dimension implied by self.loi.
        Cumulative sum is in forward order by default, but if loi integration dim starts with '-',
            instead reverse order, then cumsum, then reverse again to restore original order.

        do_interp: bool
            whether to do the interpolation step (if any) implied by self.loi.
            (disable if providing array which is already the result of loi_interp().)
        sort: UNSET, bool, 'ascending', or 'descending'
            if UNSET, use sort = self.loi_cumsum_sort
            {_sort_doc}

        See help(type(self).loi) for more details.
        '''
        if do_interp:
            array = self.loi_interp(array)
        idim = self._loi_integration_dim(signed=True)
        neg = idim.startswith('-')
        if neg:
            idim = idim[1:]
        if sort is UNSET:
            sort = self.loi_cumsum_sort
        if sort in (True, 'ascending'):
            array = array.sortby(idim)
        elif sort == 'descending':
            array = array.sortby(idim, ascending=False)
        if neg:
            reverse = {idim: slice(None,None,-1)}
            result = array.isel(reverse).cumsum(idim).isel(reverse)
        else:
            result = array.cumsum(idim)
        return result

    # [TODO] integrate-along-a-line patterns could go here.

    cls_behavior_attrs.register('loi_ustr', default=UNSET)
    loi_ustr = simple_property('_loi_ustr', default=UNSET,
        doc='''units string (e.g., 'length') for loi interped coords,
        in case self.units != self.coords_units, and loi not just along a maindim.

        Probably want to use 'length', but 'length' not assumed by LoiLoader,
            because maindims can technically represent anything, not just physical length.
        Subclass can alter default value used by editing _LOI_USTR_DEFAULT.
            (So, if subclass knows maindims and loi are lengths, it should set _LOI_USTR_DEFAULT='length')

        UNSET --> internally uses self._loi_ustr_default (default in LoiLoader: None. Subclass may override.)
        None --> if any units ambiguity (e.g. when getting dloi), crash.
        str --> use this string to get unit conversion factors as needed.
            e.g., umul = self.u(self.loi_ustr, units=self.units, convert_from=self.coords_units_explicit)
                dloi = xarray_d_grid(interped, xc) * umul   # (when loi not in maindims)''')

    loi_ustr_explicit = alias('loi_ustr',
        doc='''loi_ustr, but returns self._loi_ustr_default if loi_ustr is UNSET''')
    @loi_ustr_explicit.getter
    def loi_ustr_explicit(self):
        return self._loi_ustr_default if self.loi_ustr is UNSET else self.loi_ustr

    _loi_ustr_default = None  # subclass can override if it knows what the default should be, e.g. 'length'.

    @known_var
    def get_dloi(self):
        '''dloi = coord spacing to use for integration as implied by self.loi.
        E.g. if self.loi implies integration along 'x' then this is basically dx
            (accounting for self.slices as needed; see also: self.get_maindims_dim_didx)

        Result's values use self.units; coords use self.coords_units (default==self.units).
            (e.g., units='cgs', coord_units='si' --> values are in cgs, but coords are in SI.)

        See also: xarray_d_grid
        '''
        x = self._loi_integration_dim()  # e.g. 'x' if 'component' mode and self.component=='x'.
        # (also, e.g., 'ray_s_dim' or 'ray_s' if in 'ray_draping' mode.)
        xc = self._loi_integration_coord()  # e.g. 'x' if 'x' above
        # (also, e.g., 'ray_s' if in 'ray_draping' mode.)

        if not hasattr(self, 'maindims'):
            raise NotImplementedError('[TODO] get_dloi() for self without maindims')
            # ^ theoretically should be allowed, as long as d{x} is defined,
            #  for the appropriate x, for the relevant subclass...
            #  and, e.g., might be possible if refactoring below,
            #    to avoid relying on zeros_like_maindims().
            #  This might become necessary e.g. for FromDatasetCalculator.
        # else, self.maindims exists
        if x in self.maindims:
            result = self(f'maindims_{x}_dslice')
        else:
            if self.units != self.coords_units_explicit and (self.loi_ustr is None):
                errmsg = (f'self.loi implies integration along {x} coord,\n'
                          f'but self.coords_units ({self.coords_units!r}) != self.units ({self.units!r}),\n'
                          'and self.loi_ustr is None, so cannot reliably get dloi with proper units.\n'
                          "Consider providing self.loi_ustr, e.g. self.loi_ustr='length'.")
                raise FormulaMissingError(errmsg)
            zeros = self.zeros_like_maindims() 
            interped = self.loi_interp(zeros)
            # [TODO][EFF] maybe more efficient to not fully interpolate the array;
            #  might rewrite if this part is a bottleneck. But, in the meantime,
            #  it should give correct answer, so it is probably good enough for now.
            result = xarray_d_grid(interped, xc)
            if self.units != self.coords_units_explicit:
                umul = self.u(self.loi_ustr_explicit, units=self.units, convert_from=self.coords_units_explicit)
                # e.g. umul = self.u('length', 'cgs', 'si') == 100, when user wants cgs values but si coord units.
                result = result * umul
        return result



"""
File Purpose: sqlite3 tests for registry in PlasmaCalcs.
"""

import os
import pytest

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)
TESTS_DIR = os.path.dirname(HERE)
ARTIFACTS = os.path.join(TESTS_DIR, '_test_artifacts', 'registry_sqlite')
if os.path.exists(ARTIFACTS):  # remove any xarray_io artifacts before running tests.
    import shutil; shutil.rmtree(ARTIFACTS)
# <-- don't makedirs here; register() should makedirs as needed.


def get_calculator(name, **kw_init):
    '''get one of the standard "test-suite" calculators used in this file.
    These are:
        'ec_tiny': EppicCalculator from tests/test_eppic/test_eppic_tinybox
        'bc_tiny': BifrostCalculator from tests/test_bifrost/test_bifrost_tinybox
        'bm_tiny': BifrostMultifluidCalculator from tests/test_bifrost/test_bifrost_tinybox
    '''
    if name == 'ec_tiny':
        with pc.InDir(os.path.join(TESTS_DIR, 'test_eppic', 'test_eppic_tinybox')):
            return pc.EppicCalculator.from_here(**kw_init)
    elif name == 'bc_tiny':
        with pc.InDir(os.path.join(TESTS_DIR, 'test_bifrost', 'test_bifrost_tinybox')):
            return pc.BifrostCalculator(**kw_init)
    elif name == 'bm_tiny':
        with pc.InDir(os.path.join(TESTS_DIR, 'test_bifrost', 'test_bifrost_tinybox')):
            return pc.BifrostMultifluidCalculator(**kw_init)
    else:
        raise pc.InputError(f'unknown calculator name: {name}')


### --------------------- basic tests --------------------- ###

def test_registry_basic():
    '''test basic registry functionality, e.g. create a registry, register run'''
    ec = get_calculator('ec_tiny')
    bc = get_calculator('bc_tiny')
    bm = get_calculator('bm_tiny')

    ## in-memory registry lo register a single run
    registry00 = pc.Registry(':memory:')
    with registry00.connect() as con:
        # check column names match what we expect:
        runs_cols = [row[1] for row in con.execute('PRAGMA table_info(runs)').fetchall()]
        assert set(runs_cols) >= {'run_id', 'pc_uuid', 'calculator_type', 'registered_at_first', 'registered_at'}
        params_cols = [row[1] for row in con.execute('PRAGMA table_info(params)').fetchall()]
        assert set(params_cols) >= {'run_id', 'param_id', 'name', 'value', 'value_text'}
        # ... could repeat for other tables too, but that's a lot of stuff to hard-code,
            # which might get annoying to update here if the schema ever changes.
        # check that all tables we expect to exist actually exist:
        tables = [row[0] for row in con.execute('SELECT name FROM sqlite_master WHERE type="table"').fetchall()]
        assert set(tables) >= {'runs', 'params', 'metrics', 'register_history'}
        # register a run to this registry:
        (reg, _r00) = ec.register(registry00)
        assert _r00 is registry00
        assert len(reg) == 2   # (run_id, history_id)
        assert reg.history_id is not None   # i.e., actually added a new run.
        run_id = reg.run_id
    # check that the run we just registered is actually in the registry
    # (this is mostly a check that the ":memory:" connection logic works as expected;
    #  i.e. the registry00 connection lives if the registry00 object lives.)
    with registry00.connect() as con:
        assert (run_id,) in con.execute('SELECT run_id FROM runs').fetchall()
        assert 1 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # 1 entry
        assert 1 == registry00.sizes['runs']  # check that sizes works properly
    # check that if we make a new registry in ":memory:" it is not the same registry
    # (this is another check that the special ":memory:" processing works properly
    registry01 = pc.Registry(':memory:')
    with registry01.connect() as con:
        assert (run_id,) not in con.execute('SELECT run_id FROM runs').fetchall()
        assert 0 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # 0 entries
    # check that we can make a (different) registry in ":memory:" from a calculator
    reg, registry02 = bc.register(None)
    assert registry02.location == ':memory:'   # None <--> ":memory:"
    with registry02.connect() as con:
        assert 1 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # 1 entry
        assert 'BifrostCalculator' == con.execute('SELECT calculator_type FROM runs WHERE run_id=?',
                                                  (reg.run_id,)).fetchone()[0]

    ## registry in a file:
    REG10_PATH = os.path.join(ARTIFACTS, 'registry10.db')
    # check can register to file:
    (ec_run_id, ec_history_id), registry10 = ec.register(REG10_PATH)
    with registry10.connect() as con:
        assert (ec_run_id,) in con.execute('SELECT run_id FROM runs').fetchall()
        assert 1 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # 1 entry
    # check can register another run:
    (bc_run_id, _), _ = bc.register(registry10)
    with registry10.connect() as con:
        assert 2 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # 2 entries

    # check can register a run to same location by provided path:
    # (start with clean registry)
    REG11_PATH = os.path.join(ARTIFACTS, 'registry11.db')
    _, registry11 = ec.register(REG11_PATH)
    _, registry11b = bc.register(REG11_PATH)
    assert registry11b is not registry11  # different objects
    assert registry11b.location == registry11.location  # same file location

    ## misc check: ensure pc_uuid works as expected, i.e. reads from file when it exists:
    EC_TINYBOX_UUID = "c51a5543c96d4a41b456228bec4f9f22"   # value in tests/test_eppic/test_eppic_tinybox/_pc_uuid.txt
    with registry10.connect() as con:
        ec_uuid = con.execute('SELECT pc_uuid FROM runs WHERE run_id=?', (ec_run_id,)).fetchone()[0]
        assert ec_uuid == EC_TINYBOX_UUID   # same value as in file
        bc_uuid = con.execute('SELECT pc_uuid FROM runs WHERE run_id=?', (bc_run_id,)).fetchone()[0]
        assert bc_uuid != EC_TINYBOX_UUID   # new uuids should (practically) always be unique.

    ## re-registering a run, updating metrics (re-uses some variables from above.)
    ec.register(REG10_PATH)
    with registry10.connect() as con:
        # no new runs (this run already exists)
        assert 2 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]
        # no new history entries (no new metrics / info entered)
        assert 2 == con.execute('SELECT COUNT(*) FROM register_history').fetchone()[0]
    # repeat those tests but for bc_tiny (which doesn't have _pc_uuid.txt file natively on repo)
    #  (this is mostly a test that _pc_uuid.txt file is being properly generated.)
    bc.register(REG10_PATH)
    with registry10.connect() as con:
        assert 2 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]
        assert 2 == con.execute('SELECT COUNT(*) FROM register_history').fetchone()[0]
    # okay, now add metrics:
    bc.register(REG10_PATH, metrics=['max_mod_B', {'custom_metric0': 7}])
    with registry10.connect() as con:
        assert 2 == con.execute('SELECT COUNT(*) FROM runs').fetchone()[0]  # no new runs
        assert 3 == con.execute('SELECT COUNT(*) FROM register_history').fetchone()[0]  # 1 new history entry
        assert 2 == registry10.sizes['runs']  # check that sizes works
        assert 3 == registry10.sizes['register_history']  # check that sizes works
        # check that the new metrics are actually in the metrics table:
        modB = con.execute('''
            SELECT run_id, value, is_pc_quantity
            FROM metrics WHERE name = "max_mod_B"
        ''').fetchone()
        assert modB[0] == bc_run_id
        assert modB[1] == bc('max_mod_B')
        assert modB[2]   # came from bc('max_mod_B') so is_pc_quantity should be True
        custom0 = con.execute('''
            SELECT value, is_pc_quantity
            FROM metrics WHERE name="custom_metric0"
        ''').fetchone()
        assert custom0[0] == 7
        assert not custom0[1]  # not from pc quantity, so is_pc_quantity should be False
    # adding more of the same metrics with same values changes nothing:
    bc.register(REG10_PATH, metrics=['max_mod_B'])
    assert 2 == registry10.sizes['runs']
    assert 3 == registry10.sizes['register_history']
    # updating a metric value changes something:
    (_, history_id_4), _ = bc.register(REG10_PATH, metrics=[{'custom_metric0': 42}])
    assert 4 == registry10.sizes['register_history']
    with registry10.connect() as con:
        custom0 = con.execute('''
            SELECT value, is_pc_quantity
            FROM metrics WHERE name="custom_metric0"
        ''').fetchone()
        assert custom0[0] == 42      # updated value stored in value column,
        assert not custom0[1]  # not from pc quantity, so is_pc_quantity should be False
        history_entry = con.execute('''
            SELECT updated_metadata, updated_metrics
            FROM register_history WHERE history_id=?
        ''', (history_id_4,)).fetchone()
        assert history_entry is not None
        assert history_entry[0] == ''  # [TODO] maybe should be None instead? comes from ', '.join(empty list)
        assert history_entry[1] == 'custom_metric0'
    
    ## check that registering a run with a different calculator type but same _pc_uuid.txt
    # will be treated as the same run, and update the entry appropriately.
    (bm_run_id, _), _ = bm.register(REG10_PATH)
    assert bm_run_id == bc_run_id
    assert 2 == registry10.sizes['runs']
    assert 5 == registry10.sizes['register_history']
    # check that the calculator_type was updated to the new one
    # (also checks that registry.execute() works)
    calc_type = registry10.execute('SELECT calculator_type FROM runs WHERE run_id=?', (bm_run_id,)).fetchone()[0]
    assert calc_type == 'BifrostMultifluidCalculator'


def test_registry_merge():
    '''test registry merging.
    [TODO] more comprehensive tests...
        e.g., "if some runs in both original registries, keep values from first, only."
    '''
    ec = get_calculator('ec_tiny')
    bc = get_calculator('bc_tiny')

    _, r0 = ec.register(':memory:')
    _, r1 = bc.register(':memory:')
    r2 = r1.merge(r0)

    assert r2.sizes['sources'] == 3  # source r2, and sources from r0 and r1.
    assert r2.sizes['runs'] == 2

    rid0, uuid0 = r0.execute('SELECT run_id, pc_uuid FROM runs').fetchone()
    rid1, uuid1 = r1.execute('SELECT run_id, pc_uuid FROM runs').fetchone()

    # check that the run from r0 is in r2 with same pc_uuid, and source_run_id assigned appropriately.
    assert (rid0, uuid0) in r2.execute('SELECT source_run_id, pc_uuid FROM runs').fetchall()
    # check that the run from r1 is in r2 with same pc_uuid, and source_run_id assigned appropriately.
    assert (rid1, uuid1) in r2.execute('SELECT source_run_id, pc_uuid FROM runs').fetchall()

    # ensure that source names were transferred appropriately
    oname0 = r0.execute('SELECT name FROM sources WHERE source_id = 0').fetchone()[0]
    oname1 = r1.execute('SELECT name FROM sources WHERE source_id = 0').fetchone()[0]
    oname2 = r2.execute('SELECT name FROM sources WHERE source_id = 0').fetchone()[0]

    onames = [row[0] for row in r2.execute('SELECT name FROM sources').fetchall()]

    assert set([oname0, oname1, oname2]) == set(onames)



"""
File Purpose: loading single-fluid Muram quantities related to Equation Of State (EOS).
"""
import os

from ...errors import SnapValueError
from ...mhd import MhdEosLoader
from ...tools import simple_property, alias
from ...defaults import DEFAULTS

''' --------------------- MuramEosLoader--------------------- '''

class MuramEosLoader(MhdEosLoader):
    '''single-fluid Bifrost quantities related to Equation of State (EOS): ne, T, P.

    The implementation here assumes tables available at table=self.tabin[var],
        for var='ne', 'T', or 'P', and each having a table.interp(r=r, e=e) method,
        which gives value of var in 'raw' units.
    '''

    # non-aux functionality is inherited from MhdEosLoader.
    # 'aux' functionality is implemented here: read directly from aux files.

    EOS_MODE_OPTIONS = {**MhdEosLoader.EOS_MODE_OPTIONS,
        'aux': '''read directly from aux files for eosP, eosT, and eosne.'''
    }

    def _all_eos_aux_files_exist(self):
        '''returns whether eos aux files (eosP, eosT, and eosne) exist for all snaps in self.'''
        try:
            loadable = self.directly_loadable_vars()
        except SnapValueError:
            return False  # [TODO] this is overly restrictive...
        else:
            return all(var in loadable for var in ('eosT', 'eosP', 'eosne'))

    def _default_eos_mode(self):
        '''default for how to handle "Equation of State" related variables (ne, T, P).
        (provides default value for self.eos_mode.)

        result will be 'aux' if files for 'eosT', 'eosP', and 'eosne' exist for each snap,
        else 'table' if 'tabparams.in' file exists,
        else 'ideal'.
        '''
        if self._all_eos_aux_files_exist():
            return 'aux'
        elif os.path.isfile(os.path.join(self.dirname, 'tabparams.in')):
            return 'table'
        else:
            return 'ideal'

    # tell super().get_ne to use self('ne_aux') if eos_mode=='aux':
    _EOS_MODE_TO_NE_VAR = {**MhdEosLoader._EOS_MODE_TO_NE_VAR, 'aux': 'ne_aux'}

    # tell super().get_T to use self('T_aux') if eos_mode=='aux':
    _EOS_MODE_TO_T_VAR = {**MhdEosLoader._EOS_MODE_TO_T_VAR, 'aux': 'T_aux'}

    # tell super().get_P to use self('P_aux') if eos_mode=='aux':
    _EOS_MODE_TO_P_VAR = {**MhdEosLoader._EOS_MODE_TO_P_VAR, 'aux': 'P_aux'}

    @known_var(dims=['snap'])
    def get_ne_aux(self):
        '''electron number density, from 'eosne' file.'''
        ufactor = self.u('n', convert_from='cgs')
        return self.load_maindims_var_across_dims('eosne', u=ufactor, dims=['snap'])

    @known_var(dims=['snap'])
    def get_T_aux(self):
        '''temperature, from 'eosT' file.'''
        # note: multifluid T_aux assumes same T for all fluids.
        return self.load_maindims_var_across_dims('eosT', u='K', dims=['snap'])

    @known_var(dims=['snap'])
    def get_P_aux(self):
        '''pressure, from 'eosP' file.'''
        ufactor = self.u('pressure', convert_from='cgs')
        return self.load_maindims_var_across_dims('eosP', u=ufactor, dims=['snap'])


    # # # VAR-SPECIFIC EOS MODES # # #
    # in some cases, muram has aux data for some vars but not others;
    # in that case, might want, e.g., 'aux' eos_mode for some vars but 'table' for others.

    cls_behavior_attrs.register('eos_mode_ne', default=None)
    cls_behavior_attrs.register('eos_mode_T', default=None)
    cls_behavior_attrs.register('eos_mode_P', default=None)

    eos_mode_ne = simple_property('_eos_mode_ne',  default=None,
            doc='''None or str telling the eos_mode for electron number density, e.g. 'aux'. 
            None --> ignored; just use self.eos_mode. See help(type(self).eos_mode) for details.''')
    eos_mode_T = simple_property('_eos_mode_T',  default=None,
            doc='''None or str telling the eos_mode for temperature, e.g. 'aux'. 
            None --> ignored; just use self.eos_mode. See help(type(self).eos_mode) for details.''')
    eos_mode_P = simple_property('_eos_mode_P',  default=None,
            doc='''None or str telling the eos_mode for pressure, e.g. 'aux'. 
            None --> ignored; just use self.eos_mode. See help(type(self).eos_mode) for details.''')

    eos_mode_ne_explicit = alias('eos_mode_ne',
        doc='''alias to self.eos_mode_ne if not None, else self.eos_mode.
        Probably: internal methods use this, but user only uses eos_mode_ne.''')
    @eos_mode_ne_explicit.getter
    def eos_mode_ne_explicit(self):
        result = self.eos_mode_ne
        return self.eos_mode if result is None else result

    eos_mode_T_explicit = alias('eos_mode_T',
        doc='''alias to self.eos_mode_T if not None, else self.eos_mode.
        Probably: internal methods use this, but user only uses eos_mode_T.''')
    @eos_mode_T_explicit.getter
    def eos_mode_T_explicit(self):
        result = self.eos_mode_T
        return self.eos_mode if result is None else result

    eos_mode_P_explicit = alias('eos_mode_P',
        doc='''alias to self.eos_mode_P if not None, else self.eos_mode.
        Probably: internal methods use this, but user only uses eos_mode_P.''')
    @eos_mode_P_explicit.getter
    def eos_mode_P_explicit(self):
        result = self.eos_mode_P
        return self.eos_mode if result is None else result

    @known_var(attr_deps=[('eos_mode_ne_explicit', '_EOS_MODE_TO_NE_VAR')])
    def get_ne(self):
        '''electron number density. Depends on self.eos_mode; see help(type(self).eos_mode) for details
        (To override eos_mode for just ne, can set self.eos_mode_ne instead.)
        'ideal' --> cannot get ne. Crash with TypevarNanError.
        'table' --> ne from plugging r and e into EOS lookup tables (see self.tabin).
        'aux' --> get directly from Muram aux files.
        [more options might be available (depending on subclass) -- see self.EOS_MODE_OPTIONS.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(eos_mode=self.eos_mode_ne_explicit):
            return super().get_ne()

    @known_var(attr_deps=[('eos_mode_T_explicit', '_EOS_MODE_TO_T_VAR')])
    def get_T(self):
        '''Temperature. Depends on self.eos_mode; see help(type(self).eos_mode) for details
        (To override eos_mode for just T, can set self.eos_mode_T instead.)
        'ideal' --> cannot get temperature. Crash with TypevarNanError.
        'table' --> T from plugging r and e into EOS lookup tables (see self.tabin).
        'aux' --> get directly from Muram aux files.
        [more options might be available (depending on subclass) -- see self.EOS_MODE_OPTIONS.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(eos_mode=self.eos_mode_T_explicit):
            return super().get_T()

    @known_var(attr_deps=[('eos_mode_P_explicit', '_EOS_MODE_TO_P_VAR')])
    def get_P(self):
        '''Temperature. Depends on self.eos_mode; see help(type(self).eos_mode) for details
        (To override eos_mode for just P, can set self.eos_mode_P instead.)
        'ideal' --> cannot get pressure. Crash with TypevarNanError.
        'table' --> P from plugging r and e into EOS lookup tables (see self.tabin).
        'aux' --> get directly from Muram aux files.
        [more options might be available (depending on subclass) -- see self.EOS_MODE_OPTIONS.]
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        with self.using(eos_mode=self.eos_mode_P_explicit):
            return super().get_P()



"""
File Purpose: base quantities for MuramCalculator
"""

from ...mhd import MhdBasesLoader
import xarray as xr

### --------------------- MuramBasesLoader --------------------- ###

class MuramBasesLoader(MhdBasesLoader):
    '''base quantities based on Muram output.'''
    @known_var(dims=['snap'])
    def get_r(self):
        '''mass density. (directly from Muram)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('r')
        return self.load_maindims_var_across_dims('result_prim_r', u='mass_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_u(self):
        '''velocity. (directly from Muram)
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('u')
        return self.load_maindims_var_across_dims('result_prim_u', u='speed', dims=['snap', 'component'])

    @known_var(dims=['snap'])
    def get_e(self):
        '''energy density. (directly from Muram)
        Per unit volume, e.g. the SI units would be Joules / meter^3.
        assumes single fluid mode, i.e. result corresponds to the single fluid from Muram.
        '''
        self.assert_single_fluid_mode('e')
        return self.load_maindims_var_across_dims('result_prim_e', u='energy_density', dims=['snap'])

    @known_var(dims=['snap', 'component'])
    def get_B(self):
        '''magnetic field. (directly from Muram)'''
        return self.load_maindims_var_across_dims('result_prim_b', u='b', dims=['snap', 'component'])

    @known_var
    def get_gamma(self):
        '''adiabatic index. Assumed to be 5/3.'''
        return 5/3.

    @known_pattern(r'd([xyz])')
    def get_ds(self, var, *, _match=None):
        '''dx (or dy, or dz) --> grid cell size along the x (or y, or z) axis.
        result is probably a 0D xr.DataArray.

        Caution: this is the "underlying grid cell spacing", not numerically altered by self.slices.
        See also: self.get_maindims_dim_dslice(), which is more suitable if integrating.
        '''
        x, = _match.groups()   
        dx = self.params[f'd{x}'] * self.u('length')
        return xr.DataArray(dx)

    def get_maindims_dgrid(self, dim):
        '''return grid spacing associated with dim values (dim in self.maindims).
        Result is a 1D xarray.DataArray, with coords according to self.get_maindims_coords()[dim].
        Result's values use self.units; coords use self.coords_units (default==self.units).
            (e.g., units='cgs', coord_units='si' --> values are in cgs, but coords are in SI.)

        Caution: this is the "underlying grid cell size", regardless of self.slices.
            E.g., slicing with slice(None,None,4) will NOT lead to 4 times larger result.

        For Muram, this value comes from self.params.
        '''
        self._assert_is_maindim(dim)
        ds = self(f'd{dim}')
        coord = self.get_maindims_coord_xarrays(dim)
        result = xr.ones_like(coord) * ds
        return result
